#ifndef FLAT32
//   input   ds:si=inStream, es:di=outStream
//   output  outStream[], ds:si, es:di
	.code16
#define BX	%bx
#define SI	%si
#define DI	%di
#else
//   input   esi=inStream, edi=outStream
//   output  outStream[], ds:esi, es:edi
	.code32
#define BX	%ebx
#define SI	%esi
#define DI	%edi
#endif

//  unzx0_8088.S - ZX0 decompressor for 8088 - 73 bytes - NASM
//
//  inputs:
//  * ds:si: start of compressed data
//  * es:di: start of decompression buffer
//
//  Copyright (C) 2021 Emmanuel Marty
//  ZX0 compression (c) 2021 Einar Saukas, https://github.com/einar-saukas/ZX0
//
//  This software is provided 'as-is', without any express or implied
//  warranty.  In no event will the authors be held liable for any damages
//  arising from the use of this software.
//
//  Permission is granted to anyone to use this software for any purpose,
//  including commercial applications, and to alter it and redistribute it
//  freely, subject to the following restrictions:
//
//  1. The origin of this software must not be misrepresented; you must not
//     claim that you wrote the original software. If you use this software
//     in a product, an acknowledgment in the product documentation would be
//     appreciated but is not required.
//  2. Altered source versions must be plainly marked as such, and must not be
//     misrepresented as being the original software.
//  3. This notice may not be removed or altered from any source distribution.

zx0_decompress:
        cld                     // make string operations go forward
        movb    $0x80, %al      // initialize empty bit queue
                                // plus bit to roll into carry
        stc
        sbb     BX, BX          // initialize rep-offset to 1

.literals:
#if !defined(FLAT16) && !defined(FLAT32)
        movw    $32768, %dx
        cmpw    %dx, %si        // assume 32767 literals max
        jb      .si_ok
        subw    %dx, %si
        movw    %ds, %dx
        addb    $8, %dh
        movw    %dx, %ds
.si_ok:
#endif
        call    .get_elias      // read number of literals to copy
        rep     movsb           // copy literal bytes

        addb     %al, %al       // shift bit queue, and high bit into carry
        jc      .get_offset     // if 1: read offset, if 0: rep-match

        call    .get_elias      // read rep-match length (starts at 1)

#if !defined(FLAT16OUT) && !defined(FLAT32)
        jmp     .copy_match
.fix_di:
        subw    $256, %di
        movw    %es, %dx
        addw    $16, %dx
        movw    %dx, %es
.copy_match:
        cmpw    $-32640, %di    // assume 32639 max window
        ja      .fix_di
#else
.copy_match:
#endif
        push    SI              // save si (current pointer to compressed data)
        lea     (BX,DI), SI     // point to destination in es:di + offset in bx
#ifdef ONLY8086
.copy_loop:
        movsb %es:(SI), %es:(DI) // copy matched bytes using NMOS 8088/8086 workaround.
        loop    .copy_loop
#else
        rep movsb %es:(SI), %es:(DI)  // copy matched bytes
#endif
        pop     SI              // restore si

        addb    %al, %al        // read 'literal or match' bit
        jnc     .literals       // if 0: go copy literals

.get_offset:
        movb    $0xfe, %cl      // initialize value to FEh
        call    .elias_loop     // read high byte of match offset, set carry
        incb    %cl             // obtain negative offset high byte
        je      .done           // exit if EOD marker
        
        movb    %cl, %bh        // transfer negative high byte into bh
        movw    $1, %cx         // initialize match length value to 1
        movb    (%si), %bl      // read low byte of offset + 1 bit of len
        incw    %si             // inc instruction keep carry set
                                // set high bit that is shifted into bit 15
        rcrw    $1, %bx         // shift len bit into carry/offset in place
        call    .elias_bt       // if len bit is set, no need for more
                                // else read rest of elias-encoded match length
        incw    %cx             // fix match length
        jmp     .copy_match     // go copy match

.get_elias:
        movw    $1, %cx         // initialize value to 1
.elias_loop:
        addb    %al, %al        // shift bit queue, and high bit into carry
        jnz     .got_bit        // queue not empty, bits remain
        lodsb                   // read 8 new bits
        adcb    %al, %al        // shift bit queue, and high bit into carry
.got_bit:
.elias_bt:
        jc      .got_elias      // done if control bit is 1
        addb    %al, %al        // read data bit
        adcw    %cx, %cx        // shift into cx
        jmp     .elias_loop     // keep reading
.got_elias:
.done:
        ret
